#ifndef MINIMISATION_HH
#define MINIMISATION_HH

#include <dune/common/fmatrix.hh>
#include <dune/common/fvector.hh>
#include <dune/common/stdstreams.hh>

#include <dune/fufem/arithmetic.hh>
#include <dune/fufem/interval.hh>
#include <dune/solvers/computeenergy.hh>
#include <dune/tnnmg/problem-classes/bisection.hh>

#include "mydirectionalconvexfunction.hh"

namespace Dune {
template <class Functional>
void descentMinimisation(Functional const &J,
                         typename Functional::SmallVector &x,
                         typename Functional::SmallVector const &v,
                         Bisection const &bisection) {
  using SmallVector = typename Functional::SmallVector;
  using LocalNonlinearityType = typename Functional::NonlinearityType;

  // {{{ Construct a restriction of J to the line x + t * v

  /* We have

     1/2 <A(x+tv),x+tv>-<b,x+tv> = 1/2 <Av,v> t^2 - <b-Ax,v> t + <1/2 Ax-b,x>

     since A is symmetric.
  */
  SmallVector tmp = J.b;
  Arithmetic::addProduct(tmp, -1.0, J.A, x);
  double const JRestb = tmp * v;

  MyDirectionalConvexFunction<LocalNonlinearityType> const JRest(
      2.0 * computeEnergy(J.A, v), JRestb, *J.phi, x, v);
  // }}}

  int count;
  double const stepsize = bisection.minimize(JRest, 0.0, 0.0, count);
  dverb << "Number of iterations in the bisection method: " << count
        << std::endl;
  ;

  Arithmetic::addProduct(x, stepsize, v);
}

template <class Functional>
void minimise(Functional const &J, typename Functional::SmallVector &x,
              size_t steps, Bisection const &bisection) {
  using SmallVector = typename Functional::SmallVector;

  for (size_t step = 0; step < steps; ++step) {
    SmallVector v;
    J.gradient(x, v);
    v *= -1;

    if (v.two_norm() < 1e-14) // TODO: Make controllable
      return;

    descentMinimisation(J, x, v, bisection);
  }
}
}
#endif
